#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import re

from typing import Dict, Optional, Pattern, Tuple, TypeVar

import pyspark.sql.types as T

from loguru import logger
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import col


DataType = TypeVar("DataType", bound=T.DataType)

_RE_IS_TABLE_SEP_LINE = re.compile(r"^\s*\+(?:-+\+)+$")
_RE_SCHEMA_STRUCT = re.compile(
    r"^\s+\|--\s(.+):\s(\w+)(?:\(.+\))?\s\(nullable = (true|false)\)$"
)

STRUCT_TYPES = {
    "string": T.StringType(),
    "integer": T.IntegerType(),
    "date": T.DateType(),
    "long": T.LongType(),
    # TODO #39 implement array and nested types support
    "array": T.StringType(),
    "decimal": T.DecimalType(),
    "timestamp": T.TimestampType(),
    "boolean": T.BooleanType(),
}


def is_table_separator_line(
    line: str,
    pattern: Pattern[str] = _RE_IS_TABLE_SEP_LINE,
) -> bool:
    return bool(pattern.match(line))


def is_struct_field_row(
    row: str,
    pattern: Pattern[str] = _RE_SCHEMA_STRUCT,
) -> bool:
    return bool(pattern.match(row))


def extract_row_vals(line: str) -> Tuple[str, ...]:
    return tuple(map(str.strip, line.strip().split("|")[1:-1]))


def compose_struct_field(
    row: str,
    pattern: Pattern[str] = _RE_SCHEMA_STRUCT,
    types: Dict[str, T.AtomicType] = STRUCT_TYPES,
) -> T.StructField:
    try:
        name, type_str, nullable_str = pattern.findall(row)[0]
        if type_str == "array":
            logger.warning(
                f"casting parameter {name} as string type in row {row}."
                f"currently arrays are not supported."
            )
        return T.StructField(
            name,
            types[type_str],
            nullable=(nullable_str == "true"),
        )
    except ValueError as err:
        raise ValueError(
            f"Unknown row received {row}. Expected rows should match the "
            f"regex {pattern}, i.e. `|-- CLCAA521: string (nullable = true)`"
        ) from err
    except KeyError as err:
        raise KeyError(
            f"Unknown data type received: {type_str}. The list of known types"
            f" is {', '.join(types)}."
        ) from err


def df_show_repr_to_json(show_repr: str) -> str:
    show_repr = show_repr.strip()

    lines = filter(
        lambda line: not is_table_separator_line(line),
        show_repr.split("\n"),
    )
    header = [s.strip() for s in next(lines).split("|")][1:-1]

    return json.dumps(
        [dict(zip(header, row)) for row in map(extract_row_vals, lines)]
    )


def spark_schema_repr_to_schema(schema_repr: str) -> T.StructType:
    schema_repr = schema_repr.strip()
    return T.StructType(
        [
            compose_struct_field(row)
            for row in filter(is_struct_field_row, schema_repr.split("\n"))
        ]
    )


def apply_schema_to_df(
    raw_df: DataFrame,
    schema: T.StructType,
) -> DataFrame:
    schema_map = {f.name: f.dataType for f in schema.fields}
    logger.debug(f"Mapping dtypes to the dataframe {str(raw_df)}")

    return raw_df.select(
        *(
            col(column_name).cast(schema_map[column_name])
            for column_name in raw_df.columns
        )
    )


def generate_df_from_show_repr(
    spark: SparkSession,
    show_repr: str,
    schema_repr: Optional[str] = None,
) -> DataFrame:
    """Generate spark df from df.show and df.printSchema strings."""
    data = df_show_repr_to_json(show_repr)
    schema = spark_schema_repr_to_schema(schema_repr) if schema_repr else None

    # TODO #40 understand why df becomes empty if added schema=schema option
    raw_df = spark.read.json(
        spark.sparkContext.parallelize([data]),
    )
    return (
        apply_schema_to_df(
            raw_df,
            schema=schema,
        )
        if schema
        else raw_df
    )



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List

from pramen_py.models import MetastoreTable, TableFormat


def get_metastore_table(
    table_name: str,
    tables: List[MetastoreTable],
) -> MetastoreTable:
    def filter_func(table: MetastoreTable) -> bool:
        return table.name == table_name

    try:
        return next(filter(filter_func, tables))
    except StopIteration as err:
        raise KeyError(
            f"Table {table_name} missed in the config. "
            f"Available tables are:\n"
            f"{chr(10).join(t.name for t in tables)}"
        ) from err


def get_table_format_by_value(value: str) -> TableFormat:
    if value in TableFormat._value2member_map_:
        return TableFormat(value)
    else:
        raise Exception(f"Table format {value} is not supported by pramen-py")



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import importlib
import os
import pkgutil

from datetime import date, datetime
from functools import wraps
from pathlib import Path
from time import sleep
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Optional,
    Type,
    TypeVar,
    Union,
)

from environs import Env
from loguru import logger
from pyspark.sql import SparkSession


OPS_RET = TypeVar("OPS_RET")

DATE_FMTS = {
    "yyyy-MM-dd": "%Y-%m-%d",
    "yyyyMMdd": "%Y%m%d",
}

DEFAULT_DATE_FMT = "yyyy-MM-dd"

SPARK_DEFAULT_CONFIG = {
    "spark.master": "yarn",
    "spark.submit.deployMode": "client",
    "spark.dynamicAllocation.enabled": "true",
    "spark.dynamicAllocation.maxExecutors": "32",
}


def convert_str_to_date(
    s: str,
    fmt: str = DEFAULT_DATE_FMT,
    date_fmts: Dict[str, str] = DATE_FMTS,
) -> date:
    try:
        return datetime.strptime(s, date_fmts[fmt]).date()
    except KeyError:
        raise KeyError(
            f"Unknown date format {fmt}."
            f"The list of known formats are: "
            f"{', '.join(DATE_FMTS)}"
        )
    except ValueError:
        raise ValueError(f"Can't convert {s} to the {fmt}")


def convert_date_to_str(
    d: date,
    fmt: str = DEFAULT_DATE_FMT,
) -> str:
    try:
        return d.strftime(DATE_FMTS[fmt])
    except KeyError:
        raise KeyError(
            f"Unknown date format {fmt}."
            f"The list of known formats are: "
            f"{', '.join(DATE_FMTS)}"
        )


def load_modules(d: Path) -> None:
    """Load python modules recursively."""

    def inner(package: str) -> None:
        package_mod = importlib.import_module(package)
        for _, name, is_pkg in pkgutil.walk_packages(package_mod.__path__):
            full_name = package_mod.__name__ + "." + name
            logger.debug(
                f"Importing module {full_name} from {package_mod.__path__}"
            )
            importlib.import_module(full_name)
            if is_pkg:
                inner(full_name)

    inner(d.as_posix())


def coro(
    f: Callable[
        ...,
        Coroutine[Any, Any, OPS_RET],
    ]
) -> Callable[..., OPS_RET]:
    """Wrap coroutine into synchronous function.

    Run it with the current event loop.

    click lib can't work with the coroutines.
    """

    @wraps(f)
    def wrapper(
        *args: object,
        **kwargs: object,
    ) -> OPS_RET:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


def get_or_create_spark_session(
    env: Env,
    spark_config: Optional[Dict[str, str]] = None,
    force_recreate: bool = False,
) -> SparkSession:
    """Create or return existing SparkSession.

    In case force_recreate=True, the current session will be
    destroyed and a new one will be created.

    spark_config takes precedence over config defined via
    environment variable PRAMENPY_SPARK_CONFIG
    """
    logger.info("Preparing SparkSession")

    if env.str("PRAMENPY_SPARK_JAVA_HOME", ""):
        os.environ.update({"JAVA_HOME": env.str("PRAMENPY_SPARK_JAVA_HOME")})
        logger.info(
            "JAVA_HOME is set to " + env.str("PRAMENPY_SPARK_JAVA_HOME")
        )

    if force_recreate:
        logger.info("Force recreating a spark session")
        spark = (
            SparkSession.getActiveSession()
            if callable(getattr(SparkSession, "getActiveSession", None))
            else SparkSession.getOrCreate()  # type: ignore
        )
        logger.info("Stopping active session and waiting for 3s...")
        spark.sparkContext.stop()
        spark.stop()
        sleep(3)

    session = SparkSession.builder.config(
        "spark.app.name",
        "pramen-py",
    )
    selected_config = spark_config or env.dict(
        "PRAMENPY_SPARK_CONFIG",
        SPARK_DEFAULT_CONFIG,
    )
    for k, v in selected_config.items():
        logger.debug(f"Setting spark config option: {k} to {v}")
        session.config(k, v)

    spark = session.getOrCreate()
    logger.info("SparkSession is ready")
    return spark  # type: ignore


T_FN_ARGS = TypeVar("T_FN_ARGS")
T_FN_KWARGS = TypeVar("T_FN_KWARGS")
T_FN_RETURN = TypeVar("T_FN_RETURN")


async def run_and_retry(
    fn: Union[
        Callable[[T_FN_ARGS, T_FN_KWARGS], T_FN_RETURN],
        Callable[[T_FN_ARGS, T_FN_KWARGS], Awaitable[T_FN_RETURN]],
    ],
    *,
    number_of_trials: int,
    exception: Type[BaseException],
    retry_hook: Callable[[], Awaitable[None]],
) -> T_FN_RETURN:
    """Run a function until success or fail after number_of_trials.

    fn: a callable (could be also async function). If it has args and kwargs,
        use functools.partial(fn, *args, **kwargs)
    number_of_trials: number of times the function will be retried if
        exception raised
    exception: exception which will be caught and processed by triggering
        teh retry_hook and rerunning the function
    retry_hook: is a coroutine function which is called if the fn function
        fails

    Example:
    >>> def some_sync_foo(a, b):
    >>>     return a + b

    >>> async def some_async_foo(a, b):
    >>>     return a + b

    >>> assert 3 == await run_and_retry(
    >>>     partial(some_sync_foo, 1, 2),
    >>>     number_of_trials=2,
    >>>     exception=ValueError,
    >>>     retry_hook=AsyncMock(),
    >>> )
    >>> assert 3 == await run_and_retry(
    >>>     partial(some_async_foo, 1, 2),
    >>>     number_of_trials=2,
    >>>     exception=ValueError,
    >>>     retry_hook=AsyncMock(),
    >>> )

    """

    async def inner(
        trial: int,
        exp: Optional[Type[BaseException]] = None,
    ) -> T_FN_RETURN:
        if trial <= 0:
            raise exp  # type: ignore
        try:
            logger.info(f"Trials last: {trial} from total {number_of_trials}")
            try:
                result = await fn()  # type: ignore
            except TypeError as err:
                if "can't be used in 'await' expression" in str(err):
                    result = fn()  # type: ignore
                else:
                    raise
        except exception as exp:
            logger.error("Trial failed")
            await retry_hook()
            return await inner(trial - 1, exp)  # type: ignore
        else:
            logger.info("Trial succeeded")
            return result  # type: ignore

    # in case of partial, fn is hidden under fn.func attrib
    logger.info(
        f"Running function {fn}. "
        f"It will be tried {number_of_trials} times until success."
    )
    with logger.contextualize(tracked_by_retry=True):
        assert (
            number_of_trials > 0
        ), f"can't try to run a function {number_of_trials} times."
        return await inner(number_of_trials)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import abc
import datetime

from typing import List

import attrs

from pyspark.sql import DataFrame, SparkSession

from pramen_py.models import MetastoreTable


@attrs.define(auto_attribs=True, slots=True)
class MetastoreWriterBase(metaclass=abc.ABCMeta):
    """Adds writing capabilities to the Metastore.

    All implementations of the MetastoreWriter should be inherited
    from this base class

    Makes it possible to refer tables based on it name and configurations.
    """

    spark: SparkSession = attrs.field()
    tables: List[MetastoreTable] = attrs.field()
    info_date: datetime.date = attrs.field()

    @tables.validator
    def check_tables(  # type: ignore
        self,
        _,
        value: List[MetastoreTable],
    ) -> None:
        for table in value:
            assert isinstance(table, MetastoreTable)

    @abc.abstractmethod
    def write(
        self,
        table_name: str,
        df: DataFrame,
    ) -> None:
        """Write the given table in accordance with the configuration."""

        raise NotImplementedError



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio

import click

from click import Context, Option
from loguru import logger

from pramen_py import Transformation
from pramen_py.app import env
from pramen_py.app.logger import setup_logger
from pramen_py.runner.runner_transformation import TransformationsRunner
from pramen_py.utils import coro


CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "auto_envvar_prefix": "PRAMENPY",
}


def set_logger_callback(
    _: Context,
    __: Option,
    value: bool,
) -> bool:
    """Set debug on --verbose option.

    $ pramen-py --verbose ...
    """
    if value:
        logger.remove()
        setup_logger("DEBUG", env)

    logger.debug("DEBUG is on")
    return value


class RunGroup(click.Group):
    """Group of available transformations commands.

    We have to extend from click.Group in order to load at click.Group
    initialization available transformations dynamically from the
    transformations namespace packages.

    At this point all available transformations are initialized
    """

    def __init__(self, *args, **kwargs):  # type: ignore
        super().__init__(*args, **kwargs)
        TransformationsRunner(cli=self).activate()


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    is_eager=True,
    callback=set_logger_callback,
)
@click.version_option()
@click.pass_context
@coro
async def main(ctx: Context, verbose: bool) -> None:
    """Defining transformations for pramen in python.

    Usage examples:

    $ pramen-py transformations run
     ExampleTransformation1
     --info-date 2022-03-01
     --config tests/resources/real_config.yaml

    Application is configured via environment variables.

    $ pramen-py list-configuration-options
    # to see available options
    """
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose


@main.group()
@click.pass_context
@coro
async def transformations(
    ctx: Context,
) -> None:
    """Defines available actions for transformations."""


@transformations.command()
@click.pass_context
@coro
async def ls(ctx: Context) -> None:
    """List available transformations."""
    for t_name in set(  # noqa: C401
        t.__name__ for t in Transformation.__subclasses__()
    ):
        click.echo(t_name)


@transformations.group(cls=RunGroup)
@click.pass_context
@coro
async def run(ctx: Context) -> None:
    """Run transformations."""


@main.group()
@coro
async def completions() -> None:
    """Get shell completions scripts.

    Use --help to see available shells.

    Example:
        $ source <(pramen-py completions zsh)
    """


@completions.command(help="zsh completions script")
@coro
async def zsh() -> None:
    proc = await asyncio.subprocess.create_subprocess_shell(
        "_pramen_py_COMPLETE=zsh_source pramen-py",
        stdout=asyncio.subprocess.PIPE,
    )
    res, _ = await proc.communicate()
    click.echo(res.decode())


@completions.command(help="bash completions script")
@coro
async def bash() -> None:
    proc = await asyncio.subprocess.create_subprocess_shell(
        "_pramen_py_COMPLETE=bash_source pramen-py",
        stdout=asyncio.subprocess.PIPE,
    )
    res, _ = await proc.communicate()
    click.echo(res.decode())


@main.command()
@coro
async def list_configuration_options() -> None:
    """List of available environment variables to configure the app."""

    async def example_coroutine_function() -> str:
        with open(".env.example") as config_options:
            read_config_options = config_options.read()
        return read_config_options

    example_config_options = await example_coroutine_function()

    click.echo(example_config_options)



#  Copyright 2022 ABSA Group Limited
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pathlib
import uuid

from functools import partial
from typing import Awaitable, Callable, Generator, Optional, Type

import pytest

from _pytest.logging import LogCaptureFixture
from environs import Env
from loguru import logger
from pyspark.sql import DataFrame, SparkSession
from pytest_mock import MockerFixture

from pramen_py import Transformation
from pramen_py.app import setup_logger
from pramen_py.models import TransformationConfig
from pramen_py.runner.runner_transformation import (
    T_TRANSFORMATION,
    TransformationsRunner,
)
from pramen_py.test_utils.spark_utils import generate_df_from_show_repr
from pramen_py.utils import get_or_create_spark_session


env = Env(expand_vars=True)
env.read_env(
    path=(pathlib.Path(".") / ".env").as_posix(),
    override=True,
)

logger.remove()
setup_logger(env=env)


@pytest.fixture
def caplog(
    caplog: LogCaptureFixture,
) -> Generator[LogCaptureFixture, None, None,]:
    """Adopt caplog to work with loguru."""
    handler_id = logger.add(caplog.handler, format="{message}")
    yield caplog
    logger.remove(handler_id)


@pytest.fixture()
def mocker(mocker: MockerFixture) -> MockerFixture:
    """Mocker fixture.

    Do nothing here except exposing typing info explicitly. Otherwise, type of
    mocker was not inferred properly.
    """
    return mocker


@pytest.fixture()
def tmp_path_builder(tmp_path: pathlib.Path) -> Callable[[], pathlib.Path]:
    """Provides a factory of the random path.

    Based on pytest tmp_path, so at the end of the test session the directory
    """

    def inner() -> pathlib.Path:
        return tmp_path / str(uuid.uuid4())

    return inner


@pytest.fixture()
def spark() -> SparkSession:
    return get_or_create_spark_session(env)


@pytest.fixture()
def transformer_runner(
    spark: SparkSession,
    mocker: MockerFixture,
) -> Callable[
    [
        Type[Transformation],
        TransformationConfig,
        Optional[str],
    ],
    Awaitable[None],
]:
    """Runner of the transformations with the given config and info_date.

    The metastore file system sets to local (PRAMENPY_DEFAULT_FS=local)

    Usage example:
    >>> await transformer_runner(
    >>>     IdentityTransformer,
    >>>     config,
    >>>     "2022-04-07",   # in "yyyy-MM-dd" format
    >>> )

    >>> # table_out_path is path defined in the config for output table
    >>> result_df = spark.read.parquet(table_out_path.as_posix())
    """

    async def inner(
        transformer_cls: Type[T_TRANSFORMATION],
        config: TransformationConfig,
        info_date: Optional[str] = None,
    ) -> None:
        runner = TransformationsRunner(
            cli=mocker.Mock(),
            spark=spark,
        )
        await runner.process_transformer(
            T=transformer_cls,
            config=config,
            info_date=info_date,
        )

    return inner


@pytest.fixture()
def generate_df(spark: SparkSession) -> Callable[[str, str], DataFrame]:
    """Generate spark df from df.show and df.printSchema strings."""
    return partial(generate_df_from_show_repr, spark)
